# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 底层编码转换程序
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    读取 Awaken 脚本时, 该程序会将其文本编译成一种对于执行处理器来说更加底层的编码;
    这样做的目的在于节约重复读取脚本时的性能开支, 以及利于后续的解析执行。

    底层编码会以文件的格式保存在当前自动化项目的 Data/BaseCode 目录中;
    底层编码文件会根据脚本文件的命名与最后更新时间按照一定规则生成底层编码命名;
    读取 Awaken 脚本时, 如果存在底层编码文件则读取, 否则才会创建新的底层编码文件。

"""
import string
from pathlib import Path

from ...baseic.const import CONST
from ...baseic.common import md5_encrypt
from ...baseic.error import AwakenAnalysisBaseCodeError
from ...kit.common import ENGINE_TYPE_FUNCTION_MAP
from ...kit.global_method_map import GLOBAL_METHOD_FUNCTION_MAP


class BaseCodeConverter: ...


class BaseCodeConverter:
    """
    [ 底层编码转换程序 ]

    ---
    描述:
        用以解读 Awaken 脚本。

    """

    _temp_task_type: str
    """ 寄存器::任务类型 """
    _temp_translations: list
    """ 寄存器::译文列表 """
    _temp_decorators: list
    """ 寄存器::装饰器列表 """
    _temp_action_scope: str
    """ 寄存器::当前作用域 """
    _temp_case_is_assert: bool
    """ 寄存器::当前用例是否断言 """
    _temp_case_name: str = None
    """ 寄存器::当前用例名称 """
    _temp_codeline: str = None
    """ 寄存器::当前代码文本 """
    _temp_codeline_type: str = None
    """ 寄存器::当前代码类型 """
    _temp_codeline_number: int = None
    """ 寄存器::当前代码行数 """


    def convert(self, script_file_path: str | Path) -> tuple[str, list[str]]:
        """
        [ 转换 ]

        ---
        描述:
            解析Awaken脚本文件, 将其编译为底层编码文件。
        
        ---
        参数:
            script_file_path { str | Path } : Awaken脚本文件路径对象。

        ---
        返回:
            tuple[str, list[str]] : 任务类型 & 底层编码列表
        
        """
        self._temp_translations   = []
        self._temp_decorators     = []
        self._temp_action_scope   = CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE
        self._temp_case_is_assert = True

        script_fpath = script_file_path if isinstance(script_file_path, Path) else Path(script_file_path)
        self._temp_task_type = script_fpath.suffix.split('-')[1].upper()

        # --------------------------------------------------------------------
        # 创建脚本文件对应的底层编码文件路径对象
        # 命名规则: 脚本名-(脚本最后更新时间MD5码)-(脚本文件后缀)-basecode
        #
        # 底层编码文件不存在则尝试清理同名的历史底层编码文件, 然后重新编译脚本
        # --------------------------------------------------------------------
        basecode_fpath = CONST.Path.DirPath.BaseCode.joinpath(''.join([
            script_fpath.stem, 
            CONST.Interpreter.SYMBOL_BASECODE_FILE_SEPARATOR,
            md5_encrypt(str(script_fpath.stat().st_mtime)),
            script_fpath.suffix,
            CONST.Interpreter.SYMBOL_BASECODE_FILE_SEPARATOR,
            CONST.Interpreter.KEYWORD_BASECODE_FILE_SUFFIX
        ]))

        if not basecode_fpath.exists():
            for i in CONST.Path.DirPath.BaseCode.glob(f'{ script_fpath.stem }*'):
                if i.stem.split(CONST.Interpreter.SYMBOL_BASECODE_FILE_SEPARATOR)[0] == script_fpath.stem:
                    i.unlink()
                    break
            basecode_fpath.touch(0o777)
            try:
                self._parsing_script_and_write_codeline_file(script_fpath, basecode_fpath)
            except AwakenAnalysisBaseCodeError:
                basecode_fpath.unlink()
                raise

        # 读取底层编码文件并处理成列表
        translations = basecode_fpath.read_text(encoding='UTF-8')
        translations = [codeline for codeline in translations.split('\n') if codeline != '']
        return self._temp_task_type, translations


    def _parsing_script_and_write_codeline_file(self, script_fpath: Path, basecode_fpath: Path) -> None:
        """
        [ 解析脚本并写入文件 ]

        ---
        描述:
            NULL

        ---
        参数:
            script_fpath   { Path } : Awaken 脚本文件路径对象
            basecode_fpath { Path } : 底层编码文件路径对象
        
        """
        # 读取脚本文件
        with open(script_fpath, 'r', encoding='UTF-8') as f:
            content = f.readlines()

        self._temp_codeline_number = 1
        for codeline in content:

            # 清除换行符与注释内容
            codeline = codeline.strip().replace('\n', '')
            try:
                index = codeline.index(CONST.Interpreter.GrammarSymbol.Annotation)
            except ValueError:
                ...
            else: 
                codeline = codeline[0:index]

            # 代码文本不为空则解析
            if len(codeline) > 1:
                self._temp_codeline = codeline
                self._parse_currently_statement()

            self._temp_codeline_number += 1

        if not self._temp_case_is_assert:
            raise AwakenAnalysisBaseCodeError(CONST.Error.Interpreter.GrammarUseCaseNoAssertion.replace('#CASE_NAME#', self._temp_case_name))

        # 写入底层编码文件
        with basecode_fpath.open('a', encoding='UTF-8') as f:
            for translation in self._temp_translations:
                f.write(''.join([translation, '\n']))
                
        self._temp_translations.clear()


    def _parse_currently_statement(self) -> None:
        """
        [ 解析当前语句 ]

        ---
        描述:
            NULL

        """
        # 拆解关键字, 如果只有一个关键字则将其解析为执行语句
        keywords = [keyword for keyword in self._temp_codeline.split(' ') if keyword != '']
        if len(keywords) == 1:
            self._parse_logic_running()

        # 解析当前语句中的首个符号
        # 根据符号执行对应的解析逻辑
        else:
            symbol = self._parse_symbol(self._temp_codeline)

            try:
                {
                    # 赋值逻辑
                    CONST.Interpreter.GrammarSymbol.Give: self._parse_logic_give,

                    # 声明逻辑
                    CONST.Interpreter.GrammarSymbol.Statement: self._parse_logic_statement,

                    # 执行逻辑
                    CONST.Interpreter.GrammarSymbol.Run: self._parse_logic_running,

                    # 断言逻辑
                    CONST.Interpreter.GrammarSymbol.Assert: self._parse_logic_assert,
                    
                }[symbol]()
            
            except KeyError:
                raise AwakenAnalysisBaseCodeError(CONST.Error.Interpreter.GrammarSymbolWrongful.replace('#SYMBOL#', symbol))


    def _parse_logic_give(self):
        """
        [ 解析逻辑 :: 赋值 ]

        ---
        描述:
            NULL

        """
        # 拆解关键字
        kkey, kvalue = [keyword.replace(' ', '') for keyword in self._temp_codeline.split(CONST.Interpreter.GrammarSymbol.Give, 1)]

        # --------------------------------------------------------------------
        # 获取值中的方法执行符号的数量
        # 若方法执行符号为  0 则走赋值逻辑
        # 若方法执行符号为 !0 则走执行赋值逻辑
        # --------------------------------------------------------------------
        kvalue_call_symbol_count = kvalue[0:2].count(CONST.Interpreter.GrammarSymbol.Call)
        kvalue = kvalue[kvalue_call_symbol_count:]
        kvalue_split = kvalue.split(CONST.Interpreter.GrammarSymbol.Run)

        if kvalue_split[0] in [*GLOBAL_METHOD_FUNCTION_MAP.keys(), *ENGINE_TYPE_FUNCTION_MAP[self._temp_task_type].keys()]:
            self._temp_codeline_type = CONST.Interpreter.CodeLineType.RGive

            if kvalue_call_symbol_count != 0:
                function_region = CONST.Interpreter.CodeLineScopet.Global
            else:
                if self._temp_action_scope == CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE:
                    function_region = CONST.Interpreter.CodeLineScopet.Global
                else:
                    function_region = CONST.Interpreter.CodeLineScopet.Local

            function_keywords = []
            for value in kvalue_split:
                function_keywords.append(value)

            function_name = ''.join([function_region, CONST.Interpreter.GrammarSymbol.ScopePrefix, function_keywords[0]])
            kvalue = [keyword for keyword in [function_name, *function_keywords[1:]]]

        else:
            self._temp_codeline_type = CONST.Interpreter.CodeLineType.Give

        # --------------------------------------------------------------------
        # 赋值处理阶段
        # 获取键中资源引用符的数量并通过该值分析赋值作用域
        # --------------------------------------------------------------------
        kkey_quote_symbol_count = kkey[0:2].count(CONST.Interpreter.GrammarSymbol.Quote)
        give_value = []
        if not isinstance(kvalue, list):
            give_value.append(kvalue)
        else:
            for value in kvalue:
                give_value.append(value)

        if kkey_quote_symbol_count != 0:
            give_region = CONST.Interpreter.CodeLineScopet.Global
        else:
            if self._temp_action_scope == CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE:
                give_region = CONST.Interpreter.CodeLineScopet.Global
            else:
                give_region = CONST.Interpreter.CodeLineScopet.Local

        key_name = ''.join([give_region, CONST.Interpreter.GrammarSymbol.ScopePrefix, kkey[kkey_quote_symbol_count:]])
        translation = self._splice_translation([key_name, *give_value])
        self._temp_translations.append(translation)


    def _parse_logic_statement(self):
        """
        [ 解析逻辑 :: 声明 ]

        ---
        描述:
            NULL

        """
        # 拆解关键字
        kkey, kvalue = [keyword.replace(' ', '') for keyword in self._temp_codeline.split(CONST.Interpreter.GrammarSymbol.Statement, 1)]

        # --------------------------------------------------------------------
        # 校验参数与上一个用例是否完成断言
        # --------------------------------------------------------------------
        if kvalue == '':
            error_message = CONST.Error.Interpreter.GrammarStatementNoParameters.replace('#KEYWORD#', kkey)
            if kkey not in [
                CONST.Interpreter.StatementIdent.Case, 
                CONST.Interpreter.StatementIdent.Deco
            ]:
                error_message = CONST.Error.Interpreter.GrammarStatementWrongful.replace('#KEYWORD#', kkey)
            raise AwakenAnalysisBaseCodeError(error_message)

        if not self._temp_case_is_assert:
            raise AwakenAnalysisBaseCodeError(CONST.Error.Interpreter.GrammarUseCaseNoAssertion.replace('#CASE_NAME#', self._temp_case_name))
        
        # --------------------------------------------------------------------
        # 声明用例类型
        # --------------------------------------------------------------------
        if kkey == CONST.Interpreter.StatementIdent.Case:
            self._temp_codeline_type = CONST.Interpreter.CodeLineType.SCase
            kvalues = kvalue.split(CONST.Interpreter.GrammarSymbol.Statement, 1)
            self._temp_case_name = kvalues[0]
            self._temp_case_is_assert = False
            translation = self._splice_translation(kvalues)
            self._temp_translations.append(translation)

            # 如果装饰器声明语句列表不为空, 则替换作用域为用例名称并写入
            if len(self._temp_decorators) > 0:
                for decorator in self._temp_decorators:
                    self._temp_translations.append(decorator.replace(CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE, self._temp_case_name))
                self._temp_decorators.clear()

        # --------------------------------------------------------------------
        # 声明装饰器类型
        # --------------------------------------------------------------------
        elif kkey == CONST.Interpreter.StatementIdent.Deco:
            self._temp_codeline_type = CONST.Interpreter.CodeLineType.SDecorator
            kvalues = kvalue.split(CONST.Interpreter.GrammarSymbol.Statement, 1)
            translation = self._splice_translation(kvalues)
            self._temp_decorators.append(translation)


    def _parse_logic_running(self):
        """
        [ 解析逻辑 :: 执行 ]

        ---
        描述:
            NULL

        """
        self._temp_codeline_type = CONST.Interpreter.CodeLineType.Run
        
        # 拆解关键字
        kkey, *kvalue = [keyword.replace(' ', '') for keyword in self._temp_codeline.split(CONST.Interpreter.GrammarSymbol.Run)]

        # --------------------------------------------------------------------
        # 获取方法名中的方法调用符的数量
        # 若方法调用符号为  0 则使用全局方法
        # 若方法调用符号为 !0 则使用引擎方法
        # --------------------------------------------------------------------
        kkey_call_symbol_count = kkey.count(CONST.Interpreter.GrammarSymbol.Call)

        if kkey_call_symbol_count != 0:
            function_region = CONST.Interpreter.CodeLineScopet.Global
        else:
            if self._temp_action_scope == CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE:
                function_region = CONST.Interpreter.CodeLineScopet.Global
            else:
                function_region = CONST.Interpreter.CodeLineScopet.Local

        function_name = ''.join([function_region, CONST.Interpreter.GrammarSymbol.ScopePrefix, kkey[kkey_call_symbol_count:]])
        kvalue = [keyword for keyword in [function_name, *kvalue]]
        translation = self._splice_translation(kvalue)
        self._temp_translations.append(translation)


    def _parse_logic_assert(self):
        """
        [ 解析逻辑 :: 断言 ]

        ---
        描述:
            NULL

        """
        # 暂时每个 CASE 只会解析一条断言语句
        if self._temp_case_name:
            self._temp_codeline_type = CONST.Interpreter.CodeLineType.Assert
            keywords = [keyword for keyword in self._temp_codeline.split(CONST.Interpreter.GrammarSymbol.Assert, 1)]
            keywords = [keyword for keyword in keywords[1].split(' ') if keyword != '']
            translation = self._splice_translation(keywords)
            self._temp_translations.append(translation)
            self._temp_case_name = None
            self._temp_case_is_assert = True


    def _parse_symbol(self, codeline: str):
        """
        [ 解析代码行中的首个符号 ]

        ---
        描述:
            NULL

        """
        slicing = [i for i in codeline.split(' ') if i != '']
        for element in slicing:
            symbol = ''
            for str in element:
                if str not in string.punctuation: 
                    symbol = ''
                    break
                else:
                    symbol += str

            if len(symbol) != 0:
                return symbol


    def _splice_translation(self, keywords: list):
        """
        [ 拼接最终编码 ]

        ---
        描述:
            NULL
            
        """
        self._temp_action_scope = self._temp_case_name if self._temp_case_name else CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE

        if self._temp_codeline_type == CONST.Interpreter.CodeLineType.SCase:
            translation = f'{ self._temp_codeline_number } { CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE } { self._temp_codeline_type }'
        else:
            translation = f'{ self._temp_codeline_number } { self._temp_action_scope } { self._temp_codeline_type }'
        
        for key in keywords:
            translation += f' { key }'

        return translation



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 自定义异常
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import re

from .decorator import singleton_pattern


class AwakenErrorRecorder: ...


@singleton_pattern
class AwakenErrorRecorder:
    """
    [ 错误记录器 ]

    ---
    描述:
        NULL

    """

    _error_warehouse: list
    """ 错误信息仓库 """

    def __init__(self) -> None:
        self._error_warehouse = []


    def record(self, error_string: str) -> None:
        """
        [ 记录 ]

        ---
        描述:
            NULL

        """
        self._error_warehouse.append(error_string)


    def read(self) -> None:
        """
        [ 读取 ]

        ---
        描述:
            NULL

        """
        return self._error_warehouse


    def template_print(self) -> None:
        """
        [ 模板打印 ]

        ---
        描述:
            NULL

        """
        sym1 = '━'
        sym2 = '┃'
        sym3 = '┏'
        sym4 = '┗'
        sym5 = '┓'
        sym6 = '┛'
        tab_max_len = 50
        error_list = ['ERROR LIST', '       ']
        error_number = 0

        # 处理初始错误信息
        for error in self._error_warehouse:
            error_list.append(f'error {error_number} ::')
            if len(error) > tab_max_len:
                while 1:
                    error_list.append(error[0: tab_max_len-3])
                    error = error[tab_max_len-3:]
                    error_list.append(error)
                    if len(error) >= tab_max_len:
                        continue
                    else:
                        error_list.append(error[tab_max_len-3:])
                        error_number += 1
                        break
            else:
                error_list.append(error)
                error_number += 1
            
            error_list.append('       ')

        # 消除空字符
        error_list = [error.replace(' ', '') for error in error_list if error != '']

        # # 计算打印表字符长度
        # error_max_len = 0
        # for error in error_list:
        #     if len(error) > error_max_len:
        #         error_max_len = len(error)
        #         for s in error:
        #             if not u'\u4e00' <= s <= u'\u9fff':
        #                 error_max_len += 0.1
        # error_max_len = int(error_max_len)

        # 制表
        print(sym3 + sym1*(tab_max_len*2 + 3 - 20) + sym5)
        for error in error_list:
            offset = 0
            for s in error:
                if not u'\u4e00' <= s <= u'\u9fff':
                    offset += 1
            n = ' '*((tab_max_len*2 - len(error)*2 - 20) + offset)
            print(f'{sym2} {error} {n} {sym2}')
        print(sym4 + sym1*(tab_max_len*2 + 3 - 20) + sym6)


ERROR_RECORDER = AwakenErrorRecorder()
""" 错误记录器实例 """


class _AwakenBaseError(Exception):
    """ 
    [ 自定义异常基类 ]
    
    """

    def __init__(self, error: BaseException | str):
        self.err_name = re.findall(r'\[ (.*) \]', self.__doc__)[0]
        self.message = ''.join([self.err_name, ' :: ', str(error)])
        ERROR_RECORDER.record(self.message)

    def __str__(self) -> str:
        return self.message


class AwakenLogCreationError(_AwakenBaseError):
    """
    [ 日志创建异常 ]
    
    """


class AwakenLogOutputError(_AwakenBaseError):
    """
    [ 日志输出异常 ]
    
    """


class AwakenWebEngineError(_AwakenBaseError):
    """
    [ WEB引擎异常 ]
    
    """


class AwakenApiEngineError(_AwakenBaseError):
    """
    [ API引擎异常 ]
    
    """


class AwakenConvertCodelinError(_AwakenBaseError):
    """
    [ 转换代码行异常 ]
    
    """


class AwakenAnalysisBaseCodeError(_AwakenBaseError):
    """
    [ 解读底层编码异常 ]
    
    """


class AwakenTaskPretreatmentError(_AwakenBaseError):
    """
    [ 任务预处理异常 ]
    
    """

class AwakenWebApiServerError(_AwakenBaseError):
    """
    [ WebApi服务器异常 ]
    
    """


class AwakenDataBaseError(_AwakenBaseError):
    """
    [ 数据库异常 ]
    
    """


class AwakenWebEngineRunError(_AwakenBaseError):
    """
    [ WEB引擎运行异常 ]
    
    """




# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Awaken脚本语言解释器
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from pathlib import Path

from .structural import AwakenTask
from .base_code_converter import BaseCodeConverter


class GrammarParser: ...


class GrammarParser:
    """
    [ Awaken脚本语言解释器 ]

    ---
    描述:
        NULL

    """

    _base_code_converter: BaseCodeConverter = BaseCodeConverter()
    """ 底层编码转换类 """


    def parsing(self, script_file_path: str | Path):
        """
        [ 解析 ]

        ---
        描述:
            NULL

        ---
        参数:
            script_file_path { str | Path } : 脚本文件路径

        """
        task_type, basecodes = self._base_code_converter.convert(script_file_path)

        if len(basecodes) == 0:
            raise BaseException('Awaken 脚本文件为空 !')

        # print('-------')
        # for code in basecodes:
        #     print(code)
        # print('-------')

        return AwakenTask(task_type, basecodes)



# Copyright 2022. quinn.7@foxmail.com All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
[ 池运行时 ]

"""
from multiprocessing.pool import ThreadPool
from multiprocessing.dummy import Pool
from multiprocessing.dummy import Lock
from multiprocessing.dummy import Queue
from multiprocessing.dummy import Manager

from ...baseic.config import CONFIG


class PoolRuntime(object):
    """
    [ POOL 运行时 ]

    """

    _pool: ThreadPool
    """ 进程池 """
    _state: bool
    """ 运行状态 """
    _engine_max_count: int
    """ 引擎最大数量 """
    _task_max_count: int
    """ 任务最大数量 """
    _sign_engine_count: int
    """ 注册的驱动数量 """
    _engine_queue: Queue
    """ 引擎队列 """
    _task_queue: Queue
    """ 任务队列 """
    _state_lock: Lock
    """ 运行状态互斥锁 """
    _engine_queue_lock: Lock
    """ 引擎队列互斥锁 """
    _task_queue_lock: Lock
    """ 任务队列互斥锁 """


    def __init__(self, 
        engine_max_count: int = CONFIG.engine_queue_max_count, 
        task_max_count: int = CONFIG.task_queue_max_count
    ) -> None:
        """
        [ POOL 运行时 ]
        
        ---
        参数:
            engine_max_count { int } : 引擎最大数量
            task_max_count   { int } : 任务最大数量

        """
        self._engine_max_count = engine_max_count
        self._task_max_count = task_max_count

        manager: Manager = Manager()
        self._pool = Pool(self._engine_max_count + 1)
        self._state = False
        self._sign_engine_count = 0
        self._engine_queue = manager.Queue(maxsize=self._engine_max_count)
        self._task_queue = manager.Queue(maxsize=self._task_max_count)
        self._state_lock = manager.Lock()
        self._engine_queue_lock = manager.Lock()
        self._task_queue_lock = manager.Lock()


    @property
    def pool(self) -> ThreadPool:
        """
        [ 进程池 ]

        """
        return self._pool


    @property
    def state(self) -> bool:
        """
        [ 运行状态 ]

        """
        with self._state_lock:
            return self._state


    @property
    def engine_max_count(self) -> int:
        """
        [ 驱动队列最大值 ]

        """
        return self._engine_max_count


    @property
    def task_max_count(self) -> int:
        """
        [ 任务队列最大值 ]

        """
        return self._task_max_count


    @property
    def sign_driver_count(self) -> int:
        """
        [ 注册的核心数量 ]

        """
        return self._sign_engine_count


    @property
    def engine_queue(self) -> Queue:
        """
        [ 核心队列 ]

        """
        with self._engine_queue_lock:
            return self._engine_queue


    @property
    def task_queue(self) -> Queue:
        """
        [ 任务队列 ]

        """
        with self._task_queue_lock:
            return self._task_queue


    def set_state(self, state: bool):
        """
        [ 设置运行状态 ]
        
        ---
        参数:
            state { bool } : 运行状态

        """
        with self._state_lock:
            self._state = state


    def sign_engine(self):
        """
        [ 注册引擎 ]

        """
        self._sign_engine_count += 1



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::创建项目   
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..awaken_info import ProjectInfo
from ..baseic.common import create_project
from ..baseic.common import window_template_output


def instruction_create_project(argv: list):
    """
    [ 命令行指令::创建项目 ]

    ---
    参数:
        argv { list } : 参数列表

    """
    if len(argv) >= 1:
        success = []
        unsuccess = []
        for name in argv[0:]:
            result = create_project(name)
            if result:
                success.append(name)
            else:
                unsuccess.append(name)

            template = [
                '- 创建项目 -',
                None,
            ]
            if len(success) > 0:
                template.append('项目创建成功:')
                for name in success:
                    template.append(name)

            if len(unsuccess) > 0:
                if len(success) > 0:
                    template.append(None)
                template.append('项目创建失败:')
                for name in unsuccess:
                    template.append(name)

        window_template_output(f'Awaken2 - {ProjectInfo.Version}', template)

    else:
        window_template_output(
            f'Awaken2 - {ProjectInfo.Version}', 
            [
                '- 创建项目 -',
                None,
                '创建项目指令参数异常:',
                None,
                '示例:',
                '>> awaken -make project',
                '>> awaken -make project1 project2',
            ]
        )
        exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 项目信息
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


class ProjectInfo: ...


class ProjectInfo:
    """
    [ 项目信息 ]

    """
    Name     = 'awaken2'
    Version  = '0.0.5'
    Author   = 'chenjiancheng'
    Email    = 'quinn.7@foxmail.com'
    Desc     = 'Awaken 是以 "关键字驱动" 为中心思想搭建的自动化测试框架。'
    Url      = 'https://gitee.com/catcat7/awaken2'
    Keywords = ['python', 'testing', 'automation', 'testautomation', 'awaken2']



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 服务执行池
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import gc
import psutil

from flask import jsonify
from flask import Blueprint
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer

from .pool_runtime import PoolRuntime
from ...baseic.log import LOG
from ...baseic.const import CONST
from ...baseic.config import CONFIG
from ...server.api.awaken_app import AwakenApp
from ...kit.engine_agency import EngineAgency
from ...core.interpreter.grammar_parser import AwakenTask


class PerformServer: ...


_POOL_RUNTIME: PoolRuntime = PoolRuntime()
""" POOL运行时 """

# 创建驱动代理
for _ in range(CONFIG.engine_queue_max_count):
    _POOL_RUNTIME.engine_queue.put(EngineAgency())
    _POOL_RUNTIME.sign_engine()

# ----------------------------------------------------------------------------
# 临时测试依赖
# ----------------------------------------------------------------------------
from ...server.api.server_request_handler import URL_PREFIX


BLUEPRINT_AWAKEN = Blueprint('awaken', __name__)
""" [ Awaken 任务测试接口蓝图 ] """


@BLUEPRINT_AWAKEN.route(f'{ URL_PREFIX }/awaken/test', methods=['GET'])
def tasks():
    from ..interpreter.grammar_parser import GrammarParser
    gp = GrammarParser()
    task = gp.parsing('D:/Code/awaken/test/task.awaken-web')
    _POOL_RUNTIME.task_queue.put(task)

    return jsonify({
        'code': 0,
        'result': {
            'data': '测试任务已接收',
        },
        'message': 'ok',
        'type': 'success'
    })
# ----------------------------------------------------------------------------


class PerformServer(object):
    """
    [ 执行服务 ]

    ---
    描述:
        开启时, 将在本地挂起一个执行池进程与 WebAPI 服务。

    """

    _debug: bool
    """ 调试模式 """

    def running(self, port: int = 3700, debug: bool = None):
        """
        [ 运行 ]

        ---
        参数:
            port  { int } : 服务端口号, 默认为 3700
            debug { bool } : 调试模式

        """
        self._debug = debug if debug else CONFIG.debug
        _POOL_RUNTIME.pool.apply_async(func=self.monitor)
        
        app = AwakenApp()
        app.load_blueprint(BLUEPRINT_AWAKEN)
        
        http_server = HTTPServer(WSGIContainer(app.living))
        http_server.listen(port)

        if self._debug:
            LOG.info(CONST.Error.PerformPool.ServerRunningMessage.replace('#PORT#', str(port)))

        IOLoop.current().start()


    def monitor(self):
        """
        [ 监听进程 ]
        
        ---
        描述:
            监听来自 POOL 运行时的队列信息。

        """
        if self._debug:
            LOG.info(CONST.Error.PerformPool.ListeningProcessStarted)

        while True:

            # ----------------------------------------------------------------
            # 调用多进程执行任务
            # 驱动冷启动时会占用大量的 CPU 资源, 为避免因为 CPU 占用率过高导致进程启动异常
            # 每隔 X 秒检查 CPU 占用率, 只有 CPU 占用率少于 X 时才会启动进程
            # ----------------------------------------------------------------
            task = _POOL_RUNTIME.task_queue.get()
            engine = _POOL_RUNTIME.engine_queue.get()
            while True:
                if psutil.cpu_percent(CONFIG.cpu_property_time) <= CONFIG.cpu_property_ceiling:
                    _POOL_RUNTIME.pool.apply_async(func=self._emit, args=[task, engine])
                    break


    def _emit(self, task: AwakenTask, engine: EngineAgency):
        """
        [ 触发工作流 ]
        
        ---
        参数:
            task   { AwakenTask } : 任务对象
            engine { EngineAgency } : 驱动对象

        """
        # 核心执行任务
        task.start(engine)
        # 核心执行完任务后, 放回核心队列等待
        _POOL_RUNTIME.engine_queue.put(engine)
        # 强制 GC 回收垃圾避免内存溢出
        gc.collect()



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 静态执行池
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import gc
import psutil

from .pool_runtime import PoolRuntime
from ...baseic.log import LOG
from ...baseic.const import CONST
from ...baseic.config import CONFIG
from ...baseic.decorator import singleton_pattern
from ...kit.engine_agency import EngineAgency
from ...core.interpreter.grammar_parser import AwakenTask


class PerFormPool: ...


@singleton_pattern
class PerFormPool(object):
    """
    [ 静态执行池 ]

    ---
    描述:
        通常用于静态执行多进程任务。

    """

    _debug: bool
    """ 调试模式 """
    _pool_runtime: PoolRuntime
    """ 运行时 """

    def __init__(self, 
        engine_max_count: int = 3,
        task_max_count: int = 50,
        debug: bool = None
    ) -> None:
        """
        [ 静态执行池::初始化 ]

        ---
        参数:
            engine_max_count { int } : 引擎最大数量
            task_max_count   { int } : 任务最大数量
            debug           { bool } : 调试模式

        """
        self._debug = debug if debug else CONFIG.debug
        self._pool_runtime = PoolRuntime(engine_max_count, task_max_count)

        # 创建驱动代理
        for _ in range(engine_max_count):
            self._pool_runtime.engine_queue.put(EngineAgency())
            self._pool_runtime.sign_engine()


    def put_task(self, task: AwakenTask):
        """
        [ 推送到任务队列 ]
        
        ---
        参数:
            task { AwakenTask } : 任务对象

        """
        count = self._pool_runtime.task_queue.qsize()
        max_count = self._pool_runtime.task_queue.maxsize
        if max_count > count:
            self._pool_runtime.task_queue.put(task)

    def running(self):
        """
        [ 运行 ]

        ---
        描述:
            NULL

        """
        # --------------------------------------------------------------------
        # 开始一个监听循环
        # 当运行时的任务队列为空且所有代理引擎都空闲时, 跳出监听循环
        # --------------------------------------------------------------------
        while True:

            if self._pool_runtime.task_queue.qsize() == 0:
                if self._debug: 
                    LOG.info(CONST.Error.PerformPool.TaskQueueConsumedMessage)

                while True:
                    core_count = self._pool_runtime.engine_queue.qsize()
                    if core_count == self._pool_runtime.sign_driver_count:
                        if self._debug:
                            LOG.info(CONST.Error.PerformPool.TaskRunningConsumedMessage)
                        break
                break

            # ----------------------------------------------------------------
            # 调用多进程执行任务
            # 驱动冷启动时会占用大量的 CPU 资源, 为避免因为 CPU 占用率过高导致进程启动异常
            # 每隔 X 秒检查 CPU 占用率, 只有 CPU 占用率少于 X 时才会启动进程
            # ----------------------------------------------------------------
            task = self._pool_runtime.task_queue.get()
            engine = self._pool_runtime.engine_queue.get()
            while True:
                if psutil.cpu_percent(CONFIG.cpu_property_time) <= CONFIG.cpu_property_ceiling:
                    self._pool_runtime.pool.apply_async(func=self._emit, args=[task, engine])
                    break


    def _emit(self, task: AwakenTask, engine: EngineAgency):
        """
        [ 触发工作流 ]
        
        ---
        参数:
            task   { AwakenTask } : 任务对象
            engine { EngineAgency } : 驱动对象

        """
        # 核心执行任务
        engine.start(task)
        # 核心执行完任务后, 放回核心队列等待
        self._pool_runtime.engine_queue.put(engine)
        # 强制 GC 回收垃圾避免内存溢出
        gc.collect()



# Copyright 2022. quinn.7@foxmail.com All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ WEB 执行器 ]

"""
import time
import asyncio

from playwright.async_api import async_playwright

from .web_engine import WebEngine
from ...baseic.log import LOG
from ...baseic.const import CONST
from ...baseic.keyword import KEYWORD
from ...baseic.error import ERROR_RECORDER
from ...baseic.error import AwakenWebEngineRunError
from ...baseic.error import AwakenTaskPretreatmentError
from ...kit.global_method import GlobalMethod
from ...server.task_db_broker import TaskDbBroker
from ...core.interpreter.structural import AwakenTask
from ...core.interpreter.structural import AwakenCase
from ...core.interpreter.structural import AwakenCodeLine
from ...core.interpreter.common import converting_para_format


class WebRunner: ...


class WebRunner:
    """
    [ WEB 执行器 ]

    ---
    描述:
        NULL

    """
    _task: AwakenTask
    _global_method: GlobalMethod = GlobalMethod()
    _web_engine: WebEngine
    _task_db_broker: TaskDbBroker
    _use_case_in_execution: AwakenCase


    def __init__(self, task: AwakenTask) -> None:
        self._task = task
        self._task_db_broker = TaskDbBroker(CONST.Type.Task.Web)


    def start(self, *_):
        """
        [ 开始任务 ]

        ---
        描述:
            NULL

        """
        return asyncio.run(self._async_running_task())


    async def _async_running_task(self):
        """
        [ 协程执行任务 ]

        ---
        描述:
            NULL

        """
        try:
            async with async_playwright() as playwright:
                self._web_engine = WebEngine()
                self._web_engine._init_playwright(playwright)
                task_state = CONST.State.Result.Null

                # 数据库创建任务
                self._task_db_broker.task_start(
                    str(self._task.namespace[KEYWORD.Script.Namespace.TaskName]),
                    str(self._task.namespace[KEYWORD.Script.Namespace.TaskDocs])
                )

                # 启动浏览器
                await self._web_engine._browser_start(self._task.namespace[KEYWORD.Script.Decorator.BrowserType])

                # 遍历任务的用例
                for awaken_case in self._task.test_cases.values():
                    self._use_case_in_execution = awaken_case

                    # 数据库创建用例
                    case_id = self._task_db_broker.created_case(
                        str(self._use_case_in_execution.name),
                        str(self._use_case_in_execution.docs)
                    )

                    try:
                        case_state = CONST.State.Result.Null
                        case_run_step = ''
                        case_start_time = time.strftime("%H:%M:%S")
                        case_start_timestamp = int(time.time())

                        # 启动页面
                        await self._web_engine._page_start()

                        # 遍历用例的步骤
                        for awaken_codeline in self._use_case_in_execution.steps:

                            # 赋值逻辑
                            if awaken_codeline.type == CONST.Interpreter.CodeLineType.Give:
                                self._give_cases_assignment(
                                    awaken_codeline.give_region,
                                    awaken_codeline.give_name,
                                    awaken_codeline.give_value,
                                )

                            # 执行并赋值逻辑
                            elif awaken_codeline.type == CONST.Interpreter.CodeLineType.RGive:
                                result = await self._running_function(awaken_codeline)
                                self._give_cases_assignment(
                                    awaken_codeline.give_region,
                                    awaken_codeline.give_name,
                                    result,
                                )
                                
                            # 执行逻辑
                            elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Run:
                                await self._running_function(awaken_codeline)

                            # 断言逻辑
                            elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Assert:
                                assert_result = self.assert_logic(awaken_codeline.assert_value)

                                if assert_result == True:
                                    case_state = CONST.State.Result.Success
                                else:
                                    case_state = CONST.State.Result.Unsuccess

                    except BaseException as error:
                        case_state = CONST.State.Result.Error
                        case_run_step = str(error)

                    finally:
                        self._task_db_broker.update_case(
                            case_id,
                            case_state,
                            case_start_time,
                            case_start_timestamp,
                            case_run_step,
                        )

                    time.sleep(1)
                    await self._web_engine._page_close()

                self._task_db_broker.task_end(task_state)
                await self._web_engine._exit()

        except BaseException as error:
            import traceback
            LOG.error(traceback.format_exc())
            ERROR_RECORDER.template_print()


    async def _running_function(self, codeline: AwakenCodeLine):
        try:
            function_map = None
            function_engine = None
            
            if codeline.funtion_region == CONST.Interpreter.CodeLineScopet.Local:
                function_map = self._task.engine_function_map
                function_engine = self._web_engine
            else:
                function_map = self._task.global_function_map
                function_engine = self._global_method
            
            if codeline.funtion_name not in function_map.keys():
                raise AwakenWebEngineRunError('不合法的方法 !')

            new_function_value = self.parsing_method_parameters(codeline.funtion_value)

            if codeline.funtion_region == CONST.Interpreter.CodeLineScopet.Global:
                result = function_map[codeline.funtion_name](function_engine, *new_function_value)
            else:
                result = await function_map[codeline.funtion_name](function_engine, *new_function_value)
            return result

        except KeyError:
            LOG.info(f'方法集中不存在这样的方法 :: [{ codeline.funtion_name }] !')


    def _give_cases_assignment(self, give_region: str, give_name: str, give_value: str):
        """
        [ 私域赋值封装 ]

        """
        give_path_node = give_name.split(CONST.Interpreter.GrammarSymbol.Statement)
        give_path_node_number = len(give_path_node)
        current_node = None
        if give_region == CONST.Interpreter.CodeLineScopet.Local:
            current_node = self._use_case_in_execution.namespace
        else:
            current_node = self._task.namespace
        i = 1
        if give_path_node_number > 1:
            for node in give_path_node:
                if i == give_path_node_number:
                    current_node.update({node: give_value})
                else:
                    if node not in current_node.keys():
                        current_node.update({node: {}})
                    current_node = current_node[node]
                    i += 1
        else:
            current_node.update({give_name: give_value})


    def assert_logic(self, assert_value: list):
        condition_count = len(assert_value)
        assert_result = None

        if condition_count == 1:
            calculation_member = [assert_value[0]]
            assert_result = converting_para_format(self.parsing_method_parameters(calculation_member)[0])
            
        elif condition_count == 3:
            calculation_symbol = assert_value[1]
            calculation_lvalue = converting_para_format(self.parsing_method_parameters([assert_value[0]])[0])
            calculation_rvalue = converting_para_format(self.parsing_method_parameters([assert_value[2]])[0])

            try:
                if type(calculation_lvalue) != type(calculation_rvalue):
                    raise AwakenWebEngineRunError('断言异常 :: 参与断言的左值与右值类型不一致 !')

                if calculation_symbol == '==':
                    assert_result = calculation_lvalue == calculation_rvalue
                elif calculation_symbol == '>':
                    assert_result = calculation_lvalue > calculation_rvalue
                elif calculation_symbol == '>=':
                    assert_result = calculation_lvalue >= calculation_rvalue
                elif calculation_symbol == '<':
                    assert_result = calculation_lvalue < calculation_rvalue
                elif calculation_symbol == '<=':
                    assert_result = calculation_lvalue <= calculation_rvalue

            except BaseException:
                return False

        return assert_result == True


    def parsing_method_parameters(self, source_data: list):
        """
        [ 解析方法参数 ]

        """
        filter_data = []
        if len(source_data) > 0:
            for value in source_data:
                if isinstance(value, str):
                    value_quote_symbol_count = value[0:2].count(CONST.Interpreter.GrammarSymbol.Quote)
                    para = value[value_quote_symbol_count:]

                    if value_quote_symbol_count != 0:
                        current_node = self._task.namespace
                    else:
                        current_node = self._use_case_in_execution.namespace

                    para_statement_symbol_count = para.count(CONST.Interpreter.GrammarSymbol.Statement)
                    if para_statement_symbol_count != 0:
                        para_nodes = para.split(CONST.Interpreter.GrammarSymbol.Statement)

                        if len(para_nodes) > 1:
                            for node in para_nodes:
                                try:
                                    current_node = current_node[node]
                                except KeyError:
                                    AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node))
                        else:
                            try:
                                current_node = current_node[para_nodes[0]]
                            except:
                                AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node))

                        filter_data.append(current_node)

                    else:
                        # 如果 para 存在于私域命名空间中则取出, 否则视为普通字符串
                        if para in current_node.keys():
                            filter_data.append(current_node[para])
                        else:
                            filter_data.append(para)
                else:
                    filter_data.append(value)

        return filter_data



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::初始化工程   
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..awaken_info import ProjectInfo
from ..baseic.const import CONST
from ..baseic.common import window_template_output
from ..baseic.common import create_program_runtime_dependency_files


def instruction_init_engineering():
    """
    [ 命令行指令::初始化工程 ]

    """
    CONST.Path.FilePath.Init.touch(mode=0o777, exist_ok=True)
    
    window_template_output(
        f'Awaken2 - {ProjectInfo.Version}',
        [
            '- 初始化工程 -',
            None,
            '工程初始化成功 !',
            '初始化将创建或补全项目执行依赖文件.',
            None,
            '<*.ini> 文件仅作为工程根目录标识之用;',
            '这将帮助 Awaken 在执行指令时能够识别工程根目录.'
        ]
    )
    exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Awaken任务预处理程序
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from .structural import AwakenTask
from .structural import AwakenCase
from .structural import AwakenCodeLine
from ...baseic.log import LOG
from ...baseic.const import CONST
from ...baseic.keyword import KEYWORD
from ...baseic.error import ERROR_RECORDER
from ...baseic.error import AwakenTaskPretreatmentError
from ...kit.common import ENGINE_TYPE_FUNCTION_MAP
from ...kit.global_method_map import GLOBAL_METHOD_FUNCTION_MAP


class TaskPreprocessor:
    """
    [ Awaken任务预处理程序 ]

    ---
    描述:
        NULL

    """
    
    _task: AwakenTask
    """ 任务对象 """

    def pretreatment(self, task: AwakenTask):
        self._task = task
        try:
            self._init_namespace_by_task_type()
            self._running_preprocessing_instruction()
        except BaseException:
            import traceback
            LOG.error(traceback.format_exc())
            ERROR_RECORDER.template_print()

        return self._task


    def _init_namespace_by_task_type(self):
        self._task.namespace: dict = {
            KEYWORD.Script.Namespace.TaskType: None,
            KEYWORD.Script.Namespace.TaskName: None,
            KEYWORD.Script.Namespace.TaskDocs: None
        }

        self._task.test_cases: dict = {}

        if self._task.task_type == CONST.Type.Task.Web:
            self._task.namespace[KEYWORD.Script.Namespace.TaskType] = CONST.Type.Task.Web
            self._task.namespace[KEYWORD.Script.Decorator.BrowserType] = CONST.Type.Browser.Chromium

        elif self._task.task_type == CONST.Type.Task.Api:
            self._task.namespace[KEYWORD.Script.Namespace.TaskType] = CONST.Type.Task.Api


    def _running_preprocessing_instruction(self):
        """
        [ 运行预处理指令 ]

        ---
        描述:
            NULL

        """

        # --------------------------------------------------------------------
        # 解析全局&&引擎方法映射字典
        # --------------------------------------------------------------------
        self._task.global_function_map = GLOBAL_METHOD_FUNCTION_MAP
        self._task.engine_function_map = ENGINE_TYPE_FUNCTION_MAP[self._task.task_type]

        # --------------------------------------------------------------------
        # 循环全局域语句
        # 用例域语句将被存放至 test_cases 字典中等待引擎解析
        # --------------------------------------------------------------------
        for codeline in self._task.basecodes:
            awaken_codeline = AwakenCodeLine(codeline)

            # 公域语句
            if awaken_codeline.region == CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE:
                
                # 赋值逻辑
                if awaken_codeline.type == CONST.Interpreter.CodeLineType.Give:
                    self._give_global_assignment(awaken_codeline.give_name, awaken_codeline.give_value)

                # 执行并赋值逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.RGive:
                    result = self._running_common_function(awaken_codeline.funtion_name, awaken_codeline.funtion_value)
                    self._give_global_assignment(awaken_codeline.give_name, result)

                # 执行逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Run:
                    self._running_common_function(awaken_codeline.funtion_name, awaken_codeline.funtion_value)

                # 声明逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.SCase:
                    awaken_case = AwakenCase(
                        awaken_codeline.number,
                        awaken_codeline.case_name,
                        awaken_codeline.case_docs,
                    )
                    self._task.test_cases[awaken_case.name] = awaken_case

            # 私域语句
            else:

                # 装饰器逻辑
                if awaken_codeline.type == CONST.Interpreter.CodeLineType.SDecorator:
                    self._task.test_cases[awaken_codeline.region].decorator.update({awaken_codeline.decorator_key : awaken_codeline.decorator_value})

                # 执行并赋值逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.RGive:
                    self.verify_validity_method(awaken_codeline)

                # 执行逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Run:
                    self.verify_validity_method(awaken_codeline)

                # 其他语句
                else:
                    self._task.test_cases[awaken_codeline.region].steps.append(awaken_codeline)


    def _give_global_assignment(self, give_name: str, give_value: str):
        """
        [ 公域赋值封装 ]

        ---
        描述:
            NULL
        
        """
        give_path_node = give_name.split(CONST.Interpreter.GrammarSymbol.VariablePath)
        give_path_node_number = len(give_path_node)
        current_node = self._task.namespace
        i = 1
        if give_path_node_number > 1:
            for node in give_path_node:
                if i == give_path_node_number:
                    current_node.update({node: give_value})
                else:
                    if node not in current_node.keys():
                        current_node.update({node: {}})
                    current_node = current_node[node]
                    i += 1
        else:
            self._task.namespace.update({give_name : give_value})


    def _running_common_function(self, name: str, paras: list):
        """
        [ 公域运行方法封装 ]

        ---
        描述:
            NULL
        
        """
        try:
            if name not in self._task.global_function_map.keys():
                raise AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarMethodWrongful('#NAME#', name))

            # ----------------------------------------------------------------
            # 解析方法参数
            # ----------------------------------------------------------------
            new_function_value = []
            if len(paras) > 0:
                for value in paras:
                    if isinstance(value, str):
                        value_quote_symbol_count = value[0:2].count(CONST.Interpreter.GrammarSymbol.Quote)
                        para = value[value_quote_symbol_count:]
                        para_statement_symbol_count = para.count(CONST.Interpreter.GrammarSymbol.Statement)

                        if para_statement_symbol_count != 0:
                            current_node = self._task.namespace
                            para_nodes = para.split(CONST.Interpreter.GrammarSymbol.Statement)

                            if len(para_nodes) > 1:
                                for node in para_nodes:
                                    try:
                                        current_node = current_node[node]
                                    except KeyError:
                                        AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node))
                            else:
                                try:
                                    current_node = current_node[para_nodes[0]]
                                except:
                                    AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node))

                            new_function_value.append(current_node)

                        else:
                            # 如果 para 存在于全局命名空间中则取出, 否则视为普通字符串
                            if para in self._task.namespace.keys():
                                new_function_value.append(self._task.namespace[para])
                                continue
                            else:
                                new_function_value.append(para)
                                continue

                    else:
                        new_function_value.append(value)

            result = self._task.global_function_map[name](self._task.global_method, *new_function_value)
            return result

        except KeyError:
            raise AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarMethodWrongful('#NAME#', name))


    def verify_validity_method(self, awaken_codeline):
        """
        [ 校验方法的有效性 ]

        ---
        描述:
            NULL

        """
        function_map = None
        if awaken_codeline.funtion_region == CONST.Interpreter.CodeLineScopet.Global:
            function_map = self._task.global_function_map.keys()
        else:
            function_map = self._task.engine_function_map.keys()

        if awaken_codeline.funtion_name in function_map:
            self._task.test_cases[awaken_codeline.region].steps.append(awaken_codeline)
        else:
            raise AwakenTaskPretreatmentError(CONST.Error.Interpreter.GrammarMethodWrongful.replace('#LINE#', awaken_codeline.number).replace('#NAME#', awaken_codeline.funtion_name))
            